from typing import List
import pandas as pd
import ir_datasets
import gzip
import json
from pathlib import Path
from autoqrels import Labeler
from warnings import warn


class ZeroShotLabeler(Labeler):
    def __init__(self, cache_path=None):
        super().__init__()
        self._cache = {}
        self._cache_dirty = False
        self.cache_path = cache_path
        # pre-load cache from disk if enabled and it exists
        if self.cache_path is not None:
            # The format of the cache file for zero-shot labelers looks like:
            # {query_id: {unk_doc_id: score}}
            self.cache_path = Path(self.cache_path)
            if self.cache_path.exists():
                with gzip.open(self.cache_path, 'rt') as fin:
                    self._cache = json.load(fin)

    def infer_qrels(self, run: pd.DataFrame, qrels: pd.DataFrame = None) -> pd.DataFrame:
        if qrels is not None and len(qrels) > 0:
            warn("zeroshot labelers do not use qrels; they will be ignored")
        run = dict(iter(run.groupby('query_id')))
        result = []
        for query_id, run_group in run.items():
            unk_doc_ids = list(run_group['doc_id'])
            inferred_rel = self.infer_zeroshot(query_id, unk_doc_ids)
            result.extend(((query_id, did, rel, '1') for did, rel in zip(unk_doc_ids, inferred_rel))) # iteration=1 means inferred qrel
        self.flush_cache() # flush if cache is dirty
        return pd.DataFrame(result, columns=['query_id', 'doc_id', 'relevance', 'iteration'])

    def infer_zeroshot(self, query_id: str, unk_doc_ids: List[str]) -> List[float]:
        # get results from cache
        unk_doc_ids = list(unk_doc_ids)
        c = self._cache.setdefault(query_id, {})
        res = [c.get(did) for did in unk_doc_ids]

        # calculate missing values (if any)
        uncached_idxs = [idx for idx, r in enumerate(res) if r is None]
        if uncached_idxs:
            uncached_dids = [unk_doc_ids[idx] for idx in uncached_idxs]
            uncached_res = self._infer_zeroshot(query_id, uncached_dids)
            for did, idx, score in zip(uncached_dids, uncached_idxs, uncached_res):
                res[idx] = score
                c[did] = score
            # mark cache as dirty so it can be flushed later
            self._cache_dirty = True
        return res

    def flush_cache(self):
        if self._cache_dirty and self.cache_path is not None:
            with ir_datasets.util.finialized_file(str(self.cache_path), 'wb') as fout, \
                 gzip.open(fout, 'wt') as gz_fout:
                json.dump(self._cache, gz_fout)
            self._cache_dirty = False

    def _infer_zeroshot(self, query_id: str, unk_doc_ids: List[str]) -> List[float]:
        raise NotImplementedError() # implementation-specific



def get_text(dataset, ids, text_field):
    if isinstance(ids, str):
        return get_text(dataset, [ids], text_field)[0]
    ids = list(ids)
    entities = dataset.lookup(ids)
    if text_field is None:
        texts = [entities[i][1] for i in ids]
    elif isinstance(text_field, str):
        texts = [getattr(entities[i], text_field) for i in ids]
    else:
        texts = ['\n'.join([getattr(entities[i], f) for f in text_field]) for i in ids]
    return texts

def query_text(dataset, query_ids, text_field):
    return get_text(dataset.queries, query_ids, text_field)

def doc_text(dataset, doc_ids, text_field):
    return get_text(dataset.docs, doc_ids, text_field)



from functools import cached_property
from typing import List
import ir_datasets
import more_itertools
import torch
from transformers import AutoModelForSeq2SeqLM, AutoTokenizer
import smashed
import autoqrels
from . import ZeroShotLabeler


logger = ir_datasets.log.easy()


class GradedMonoPrompt(ZeroShotLabeler):
    PROMPT = """Instruction: Indicate if the passage answers the question.
###
Example 1:
Question: At about what age do adults normally begin to lose bone mass?
Passage: For most people, bone mass peaks during the third decade of life. By this age, men typically have accumulated more bone mass than women. After this point, the amount of bone in the skeleton typically begins to decline slowly as removal of old bone exceeds formation of new bone.
Answer: Perfectly relevant
###
Example 2:
Question: when and where did the battle of manassas take place
Passage: Summary of the Battle of Bull Run. The conflict took place close to Manassas Junction, Virginia. Around 35,000 Union soldiers marched from Washing D.C. towards Bull Run (a small river) where a 20,000 troop Confederate force was stationed.
Answer: Irrelevant
###
Example 3:
Question: which kind of continental boundary is formed where two plates move horizontally past one another?
Passage: One plate slides horizontally past another. The best-known example is the earthquake-prone San Andreas Fault Zone of California, which marks the boundary between the Pacific and North America Plates. See: Teaching About Transform Plate Boundaries.
Answer: Highly relevant
###
Example 4:
Question: what foods should you stay away from if you have asthma
Passage: Get early and regular prenatal care. The first 8 weeks of your pregnancy are important to your baby's development. Early and regular prenatal care can boost your chances of having a safe pregnancy and a healthy baby. Prenatal care includes screenings, regular exams, pregnancy and childbirth education, and counseling and support.
Answer: Irrelevant
###
Example 5:
Question: what is lbm in body composition
Passage: They also measured the participantsâ€™ body fat and â€œlean body massâ€ â€“ a measure. Trusted Source. of muscle mass, obtained by subtracting the body fat weight from the total body weight.
Answer: Relevant
###
Example 6:
Question: {{ query_text }}
Passage: {{ unk_doc_text }}
Answer:"""

    def __init__(self, dataset, backbone='google/flan-t5-xl', device=None, batch_size=16, verbose=False, query_field=None, doc_field=None, max_src_len=330, cache_path=None):
        super().__init__(cache_path=cache_path)
        self.backbone = backbone
        self.tokeniser = AutoTokenizer.from_pretrained(backbone)
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        self.batch_size = batch_size
        self.verbose = verbose
        self.rel_token_id = self.tokeniser("Yes", add_special_tokens=False).input_ids[0]
        self.non_rel_token_id = self.tokeniser("No", add_special_tokens=False).input_ids[0]
        self.dataset = dataset
        self.query_field = query_field
        self.doc_field = doc_field
        fields = ['query_text', 'rel_doc_text', 'unk_doc_text']
        m_jinja = smashed.mappers.JinjaMapper(jinja=self.PROMPT)
        m_txt2word = smashed.mappers.TextToWordsMapper(
            fields=fields,
        )
        prompt_length = max_src_len - len(
            m_txt2word.splitter(m_jinja.template_text[0])
        )
        self.encode_mapper = (
            m_txt2word
            >> smashed.mappers.TruncateMultipleNestedFieldsMapper(
                fields_to_truncate=fields,
                max_length=prompt_length,
            )
            >> smashed.mappers.WordsToTextMapper(
                fields=fields,
            )
            >> m_jinja
            >> smashed.mappers.TokenizerMapper(
                tokenizer=self.tokeniser,
                input_field="source",
                add_special_tokens=False, # prompt itself has a </s> at the end!
                truncation=True,
            )
            >> smashed.mappers.ChangeFieldsMapper(
                keep_fields=["input_ids", "attention_mask"],
            )
            >> smashed.mappers.Python2TorchMapper()
            >> smashed.mappers.FixedBatchSizeMapper(
                batch_size=batch_size,
            )
            >> smashed.mappers.FromTokenizerTensorCollatorMapper(
                tokenizer=self.tokeniser,
                fields_pad_ids={"labels": -100},
            )
        )
        self.decode_mapper = (
            smashed.mappers.Torch2PythonMapper()
            >> smashed.mappers.UnpackingMapper(
                fields_to_unpack=["labels", "input_ids", "prob"],
                ignored_behavior="drop",
            )
            >> smashed.mappers.DecodingMapper(
                tokenizer=self.tokeniser,
                fields=["input_ids", "labels"],
                skip_special_tokens=True,
                clean_up_tokenization_spaces=True,
            )
        )

    def __repr__(self):
        return 'autoqrels.zeroshot.GradedMonoPrompt'

    @cached_property
    def model(self):
        return AutoModelForSeq2SeqLM.from_pretrained(self.backbone).eval().to(self.device)

    def _infer_zeroshot(self, query_id: str, unk_doc_ids: List[str]) -> List[float]:
        return self.infer_zeroshot_text(
            autoqrels.text.query_text(self.dataset, query_id, self.query_field),
            autoqrels.text.doc_text(self.dataset, unk_doc_ids, self.doc_field))

    def infer_zeroshot_text(self, query_text: str, unk_doc_text: List[str]) -> List[float]:
        prompt_data = self.encode_mapper.map([
            {'query_text': query_text, 'unk_doc_text': unk_doc_text}
            for unk_doc_text in unk_doc_texts
        ])
        preds: List[Dict[str, str]] = []

        if self.verbose:
            prompt_data = logger.pbar(prompt_data, desc=repr(self), unit='batch')

        with torch.no_grad():
            for batch in prompt_data:
                input_ids = batch["input_ids"].to(self.device)
                attention_mask = batch["attention_mask"].to(self.device)
                decoder_input_ids = torch.full(
                    (input_ids.size(0), 1),
                    int(self.model.config.decoder_start_token_id),
                    dtype=torch.long,
                    device=self.device,
                )
                output = self.model(
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    decoder_input_ids=decoder_input_ids,
                )
                logits = output.logits[
                    :, -1, [self.non_rel_token_id, self.rel_token_id]
                ].to(torch.float32)
                probs = torch.softmax(logits, dim=-1)

                # decode the generated string
                batch_pred = self.decode_mapper.map(
                    [
                        {
                            **batch,
                            "labels": torch.argmax(output.logits, dim=-1),
                            "prob": probs[:, 1],
                        }
                    ]
                )
                preds.extend([r['prob'] for r in batch_pred])

        return preds



from typing import List
from . import OneShotLabeler


class MaxRep(OneShotLabeler):
    def __init__(self, corpus_graph, score_mode='recip_rank', cache_path=None):
        super().__init__(cache_path=cache_path)
        self.corpus_graph = corpus_graph
        self.score_mode = score_mode

    def __repr__(self):
        return 'autoqrels.oneshot.MaxRep'

    def _infer_oneshot(self, query_id: str, rel_doc_id: str, unk_doc_ids: List[str]) -> List[float]:
        idx_map = {did: i for i, did in enumerate(unk_doc_ids)}
        result = [0. for _ in idx_map]
        neighbours = list(self.corpus_graph.neighbours(rel_doc_id))
        for i, neighbour_id in enumerate(neighbours):
             if neighbour_id in idx_map:
                idx = idx_map[neighbour_id]
                if self.score_mode == 'recip_rank':
                    result[idx] = (len(neighbours)-i) / len(neighbours)
                else:
                    raise ValueError(f'unsupported score_mode: {repr(self.score_mode)}')
        return result



from typing import List
from functools import cached_property
import ir_datasets
import more_itertools
import torch
from transformers import T5ForConditionalGeneration, AutoTokenizer
import autoqrels
from . import OneShotLabeler


logger = ir_datasets.log.easy()


class DuoT5(OneShotLabeler):
    def __init__(self, dataset, model_name='castorini/duot5-3b-msmarco', tokeniser_name='t5-3b', device=None, batch_size=8, verbose=False, query_field=None, doc_field=None, cache_path=None):
        super().__init__(cache_path=cache_path)
        self.model_name = model_name
        self.tokeniser = AutoTokenizer.from_pretrained(tokeniser_name, model_max_length=512)
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        self.batch_size = batch_size
        self.verbose = verbose
        self.REL = self.tokeniser.encode('true')[0]
        self.NREL = self.tokeniser.encode('false')[0]
        self.dataset = dataset
        self.query_field = query_field
        self.doc_field = doc_field

    @cached_property
    def model(self):
        return T5ForConditionalGeneration.from_pretrained(self.model_name).eval().to(self.device)

    def __repr__(self):
        return 'autoqrels.oneshot.DuoT5'

    def _infer_oneshot(self, query_id: str, rel_doc_id: str, unk_doc_ids: List[str]) -> List[float]:
        return self.infer_oneshot_text(
            autoqrels.text.query_text(self.dataset, query_id, self.query_field),
            autoqrels.text.doc_text(self.dataset, rel_doc_id, self.doc_field),
            autoqrels.text.doc_text(self.dataset, unk_doc_ids, self.doc_field))

    def infer_oneshot_text(self, query_text: str, rel_doc_text: str, unk_doc_texts: List[str]) -> List[float]:
        it = ((query_text, rel_doc_text, d) for d in unk_doc_texts)
        if self.verbose:
            it = logger.pbar(list(it), desc=repr(self), unit='d')
        result = []
        with torch.no_grad():
            for chunk in more_itertools.chunked(it, self.batch_size):
                batch = self.tokeniser.batch_encode_plus(
                    [f'Query: {q} Document0: {t2} Document1: {t1} Relevant:' for q, t1, t2 in chunk],
                    return_tensors='pt',
                    padding=True,
                    truncation=True)
                batch['decoder_input_ids'] = torch.full(
                    (batch['input_ids'].shape[0], 1),
                    self.model.config.decoder_start_token_id,
                    dtype=torch.long
                )
                res = self.model(**{k: v.to(self.device) for k, v in batch.items()})
                scores = res['logits'][:, 0, [self.REL, self.NREL]]
                scores = scores.softmax(dim=1)[:, 0].cpu().tolist()
                result.extend(scores)
        return result



import pandas as pd
import ir_measures
from ir_measures.providers.cwl_eval import CwlEvaluator


SUPPORTED_MEASURES = {'SDCG', 'P', 'RBP'}


class Labeler:

    def infer_qrels(self, run: pd.DataFrame, qrels: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError() # implementation-specific

    def __getattr__(self, attr: str):
        if attr in SUPPORTED_MEASURES:
            measure = getattr(ir_measures.measures, attr)
            return self._measure_factory(measure)
        return self.__getattribute__(attr)

    def _measure_factory(self, M):
        _SUPPORTED_PARAMS = dict(M.SUPPORTED_PARAMS)
        name = repr(self) + '.' + M.NAME
        class _RuntimeMeasure(ir_measures.measures.Measure):
            nonlocal _SUPPORTED_PARAMS
            SUPPORTED_PARAMS = _SUPPORTED_PARAMS
            NAME = M.NAME
            __name__ = name
            _autoqrels_provider = self
            _autoqrels_base_measure = M

            def runtime_impl(self, qrels, run):
                inf_qrels = self._autoqrels_provider.infer_qrels(run, qrels)
                evaluator = CwlEvaluator([self], inf_qrels, {(None, 0., 1.): [self]}, verify_gains=False)
                return evaluator.iter_calc(run)
        Measure = _RuntimeMeasure()
        if 'max_rel' in _SUPPORTED_PARAMS:
            Measure = Measure(max_rel=1)
        return Measure


class DummyLabeler(Labeler):
    def infer_qrels(self, run: pd.DataFrame, qrels: pd.DataFrame) -> pd.DataFrame:
        return qrels



from functools import cached_property
from typing import List
import ir_datasets
import more_itertools
import torch
from transformers import AutoModelForSeq2SeqLM, AutoTokenizer
import smashed
import autoqrels
from . import OneShotLabeler


logger = ir_datasets.log.easy()


class DuoPrompt(OneShotLabeler):
    PROMPT = (
        "Determine if passage B is as relevant as passage A "
        "for the given query. "
        'Passage A: "...{{ rel_doc_text | replace("\\"", "\'") }}..." '
        'Passage B: "...{{ unk_doc_text | replace("\\"", "\'") }}..." '
        'Query: "{{ query_text }}" '
        "Is passage B as relevant as passage A? </s>"
    )

    def __init__(self, dataset, backbone='google/flan-t5-xl', device=None, batch_size=8, verbose=False, query_field=None, doc_field=None, max_src_len=330, cache_path=None):
        super().__init__(cache_path=cache_path)
        self.backbone = backbone
        self.tokeniser = AutoTokenizer.from_pretrained(backbone)
        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = torch.device(device)
        self.batch_size = batch_size
        self.verbose = verbose
        self.rel_token_id = self.tokeniser("Yes", add_special_tokens=False).input_ids[0]
        self.non_rel_token_id = self.tokeniser("No", add_special_tokens=False).input_ids[0]
        self.dataset = dataset
        self.query_field = query_field
        self.doc_field = doc_field
        fields = ['query_text', 'rel_doc_text', 'unk_doc_text']
        m_jinja = smashed.mappers.JinjaMapper(jinja=self.PROMPT)
        m_txt2word = smashed.mappers.TextToWordsMapper(
            fields=fields,
        )
        prompt_length = max_src_len - len(
            m_txt2word.splitter(m_jinja.template_text[0])
        )
        self.encode_mapper = (
            m_txt2word
            >> smashed.mappers.TruncateMultipleNestedFieldsMapper(
                fields_to_truncate=fields,
                max_length=prompt_length,
            )
            >> smashed.mappers.WordsToTextMapper(
                fields=fields,
            )
            >> m_jinja
            >> smashed.mappers.TokenizerMapper(
                tokenizer=self.tokeniser,
                input_field="source",
                add_special_tokens=False, # prompt itself has a </s> at the end!
                truncation=True,
            )
            >> smashed.mappers.ChangeFieldsMapper(
                keep_fields=["input_ids", "attention_mask"],
            )
            >> smashed.mappers.Python2TorchMapper()
            >> smashed.mappers.FixedBatchSizeMapper(
                batch_size=batch_size,
            )
            >> smashed.mappers.FromTokenizerTensorCollatorMapper(
                tokenizer=self.tokeniser,
                fields_pad_ids={"labels": -100},
            )
        )
        self.decode_mapper = (
            smashed.mappers.Torch2PythonMapper()
            >> smashed.mappers.UnpackingMapper(
                fields_to_unpack=["labels", "input_ids", "prob"],
                ignored_behavior="drop",
            )
            >> smashed.mappers.DecodingMapper(
                tokenizer=self.tokeniser,
                fields=["input_ids", "labels"],
                skip_special_tokens=True,
                clean_up_tokenization_spaces=True,
            )
        )

    def __repr__(self):
        return 'autoqrels.oneshot.DuoPrompt'

    @cached_property
    def model(self):
        return AutoModelForSeq2SeqLM.from_pretrained(self.backbone).eval().to(self.device)

    def _infer_oneshot(self, query_id: str, rel_doc_id: str, unk_doc_ids: List[str]) -> List[float]:
        return self.infer_oneshot_text(
            autoqrels.text.query_text(self.dataset, query_id, self.query_field),
            autoqrels.text.doc_text(self.dataset, rel_doc_id, self.doc_field),
            autoqrels.text.doc_text(self.dataset, unk_doc_ids, self.doc_field))

    def infer_oneshot_text(self, query_text: str, rel_doc_text: str, unk_doc_texts: List[str]) -> List[float]:
        prompt_data = self.encode_mapper.map([
            {'query_text': query_text, 'rel_doc_text': rel_doc_text, 'unk_doc_text': unk_doc_text}
            for unk_doc_text in unk_doc_texts
        ])
        preds: List[Dict[str, str]] = []

        if self.verbose:
            prompt_data = logger.pbar(prompt_data, desc=repr(self), unit='batch')

        with torch.no_grad():
            for batch in prompt_data:
                input_ids = batch["input_ids"].to(self.device)
                attention_mask = batch["attention_mask"].to(self.device)
                decoder_input_ids = torch.full(
                    (input_ids.size(0), 1),
                    int(self.model.config.decoder_start_token_id),
                    dtype=torch.long,
                    device=self.device,
                )
                output = self.model(
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    decoder_input_ids=decoder_input_ids,
                )
                logits = output.logits[
                    :, -1, [self.non_rel_token_id, self.rel_token_id]
                ].to(torch.float32)
                probs = torch.softmax(logits, dim=-1)

                # decode the generated string
                batch_pred = self.decode_mapper.map(
                    [
                        {
                            **batch,
                            "labels": torch.argmax(output.logits, dim=-1),
                            "prob": probs[:, 1],
                        }
                    ]
                )
                preds.extend([r['prob'] for r in batch_pred])

        return preds
